#!/bin/bash
# set -x
################################################################################
# YAP Tools: Script for Automating YAP Synthesis Experiments 
# CC-BY-NC 4.0, (c) Mario Gleirscher
#
# Note: This script is pre-configured with settings for the CSI:Cobot workcell
# example and similar examples.
################################################################################

USAGE="
Usage: ${0} [-2cdehilrstvwy] [-p N]
  
       This script automates some core steps of synthesis experiments 
       with YAP and further tools.  By default (-w), a design space is 
       generated (-y), followed by the verified synthesis of policies (-e),
       and the visualisation of a pre-selected policy (-v).  A few other 
       steps and options are available with the following options:

       -2     plot 2D Pareto front
       -a     perform (quantitative) analyses (only with or after -p)
       -c     generate controller code
       -d     generate information for debugging
       -e     (re-)run EvoChecker 
       -h     help
       -i     insert decision points (only with or after -y)
       -l     show logs
       -p N   (re-)run Prism and generate policy no. N
       -r     generate test reference model
       -s     stay silent (no windows will be opened, solely file-based output)
       -t     generate abstract traces
       -v     visualise Pareto front (with -e) and a specific policy (with -p)
       -w     run overall procedure (default variant of experiment including the
              generation of controller code from a particular optimal solution)
       -y     run YAP

System requirements: 
       Tools:
       YAP, evoChecker, PRISM, Python 3, matplotlib, libfsmtest,
       Java/OpenJDK (11+ for EvoChecker and YAP), zathura, bash, sed, 
       GraphViz (dot), g++, gnome-terminal, grep, pdfcrop, tee, zip, 
       POSIX file utils (rm, etc.)

       Libraries:
       /usr/lib/libfsmtest.a

Examples: 
       Run overall procedure, picking optimal solution no. 24:
       ./run.sh -w -p 24 

       Run YAP, convert MDP into pDTMC, run evoChecker, and visualise Pareto front:
       ./run.sh -y -i -e -v

Note:  Make sure that the example directly is clear of obsolete files when 
       not using -w.

This script is part of the YAP Tools.
"

################################################################################
# dataset (experiment settings)
TYP=workcell
MDL=${TYP}
# file with optimisation objectives (can be generated by YAP)
OBJ=cost-prod-risk-HC-HS-HRW
# process activity to assess
ACT=exchWrkp
# property file for quantitative policy/DTMC assessment
PROPS=workcell_pol.props
# ACT=${TYP}
TPL=workcell-ctr-tpl
# Model execution period, e.g. 50
MXP=75
# Size of the random search population, e.g. 50
POP=100
# Search iterations, e.g. 50
MEVAL=100

# defaults:
SOL=9
# IMPORTANT: manually choose from file 03-paretofronts/${OBJ}_set
# Pick the line number and subtract 2

################################################################################
# tool/installation settings (for Linux)
PDFVIEWER=zathura
PRISMDIR=/opt/prism
PRISM=${PRISMDIR}/bin/prism
ECDIR=/opt/EvoChecker
ECBIN=${ECDIR}/EvoChecker-1.0.1.jar
JAVABIN=/usr/lib/jvm/java-14-openjdk-amd64/bin/java
vmArgs="-Xmx3g -XX:ParallelGCThreads=1"
YAPTOOLDIR=/usr/share/yapp/tools
VISTOOLDIR=${YAPTOOLDIR}
# VISTOOLDIR=${ECDIR}/scripts
PRJDIR=${HOME}/yap-examples/hrc2
LIBFSMTESTDIR=/opt/libfsmtest

FRONT=03-paretofronts/${OBJ}_front
SET=03-paretofronts/${OBJ}_set

usage() {
    echo "${USAGE}" 1>&2
    exit 1
}

error() {
    echo "Error code ${?}"
    exit 1
}

# defaults for CLI parameters
DOANA=0
BSILENT=0
DOYAP=0
DOYAPC=0
DOEVO=0
DOIDP=0
DOLOG=0
DOPRISM=0
DOVIS=0
DOTRACES=0
DOTEST=0
DETAIL=0
DIM=3

# check and initialize parameters
while getopts "2acdelp:rstvwhiy" OPTION; do
    case ${OPTION} in
        2) DIM=2
           ;;
	a) DOANA=1
	   ;;
	c) DOYAPC=1
	   ;;
	e) DOEVO=1
	   ;;
	i) DOIDP=1
	   ;;
	l) DOLOG=1
	   ;;
	p) DOPRISM=1
           SOL=${OPTARG}
	   ;;
	r) DOTEST=1
	   ;;
	s) BSILENT=1
	   ;;
	t) DOTRACES=1
	   ;;
	v) DOVIS=1
	   ;;
        w) DOYAP=1
           DOEVO=1
           DOPRISM=1
           DOVIS=1
           DOYAPC=1
           ;;
        y) DOYAP=1
           ;;
	*) usage
	   ;;
    esac
done
# shift $((OPTIND-1))

# set to default procedure/experiment if called without parameters
if [[ $# < 1 ]]; then
    DOYAP=1
    DOEVO=1
    DOPRISM=1
    DOVIS=1
    DOYAPC=1
    BSILENT=1
fi

################################################################################
if [[ ${DOLOG} == 1 ]]; then
    echo "Show log file(s) ..."
    tail -n 200 ~/.yap/yap.log
    exit 0
fi

################################################################################
if [[ ${DOYAP} == 1 ]]; then 
    echo "1.1.1) Generate controller design space using YAP ..."
    echo "(a) pDTMC (not PRISM-compatible with mitigation decision points):"
    sed -e '/YAP#\(ENUMS\|EVOLVABLES_FIX\|PREDICATES\)%/d' <${PRJDIR}/01-process/${MDL}-tpl.pm \
        >${PRJDIR}/01-process/${MDL}-tpl-pdtmc.pm
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/01-process/${MDL}-tpl-pdtmc.pm" \
         -o "${PRJDIR}/02-designspace/${MDL}.pm" \
         --global-logging -l 2 -f evochecker -d turn-based \
         --expand-formulas --synthesise controller \
        || { echo "YAP failed." && exit; };
    # --controller-optional 
    
    echo "(b) MDP (PRISM-compatible with non-deterministic controller):"
    sed -e '/^dtmc/s/dtmc/mdp/; /evolve/d; /^\/\/ const/s/^\/\/ //; /EVOLVABLES%/d' \
        <${PRJDIR}/01-process/${MDL}-tpl.pm \
        >${PRJDIR}/01-process/${MDL}-tpl-mdp.pm
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/01-process/${MDL}-tpl-mdp.pm" \
         -o "${PRJDIR}/02-designspace/${MDL}-mdp-compact.pm" \
         --global-logging -l 2 -f evochecker -d turn-based \
         --synthesise controller --non-deterministic \
        || { echo "YAP failed." && exit 1; };

    mv ${PRJDIR}/02-designspace/*.props ${PRJDIR}/02-requirements/
fi

################################################################################
if [[ ${DOIDP} == 1 ]]; then
    echo "1.1.2) Translate MDP into pDTMC ..."
    echo "Unfolding MDP with PRISM ..."
    ${PRISM} "${PRJDIR}/02-designspace/${MDL}-mdp-compact.pm" \
	     -exportmodel 04-process-opt/${MDL}-mdp-compact.tra,sta,lab \
             -mdp >04-process-opt/${MDL}-prism.log \
        || { echo "Expansion of the MDP transition relation (using PRISM) failed." && exit 1; };

    echo "Converting MDP transition matrix into pDTMC by inserting decision points ..."
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/01-process/${MDL}-tpl-with-dps.pm" \
         -e "${PRJDIR}/04-process-opt/${MDL}-mdp-compact" \
         -o "${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm" \
         --global-logging -l 2 -f pdtmcalldps -d turn-based \
         --synthesise controller --non-deterministic \
         --expand-formulas || { echo "YAP failed." && exit 1; };

    echo "Adjusting model execution period ..."
    sed -i "s/\\(const int T = \\)[0-9]\\{1,3\\};/\\1${MXP};/" \
        ${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm \
        || { echo "Failed in adjusting model execution period." && exit 1; };
    
    echo "Resetting the initial state ..."
    # turn=2: it's the operator's turn 
    # ract=2: robot is in exchWrkp activity
    sed -i 's/\(\(turn\|ract\): \[0\.\.2\] init \)[0-9]/\12/' \
        ${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm \
        || { echo "Failed in resetting initial state." && exit 1; };

    echo "Replacing normal pDTMC with decision-point enhanced pDTMC ..."
    mv ${PRJDIR}/02-designspace/${MDL}-pdtmc-with-dps.pm ${PRJDIR}/02-designspace/${MDL}.pm
fi

################################################################################
# generate EvoChecker config file (for Linux)
ECCFG="
PROBLEM = ${TYP}
MODEL_TEMPLATE_FILE = ${PRJDIR}/02-designspace/${MDL}.pm
PROPERTIES_FILE     = ${PRJDIR}/02-requirements/${OBJ}.pctl
ALGORITHM = NSGAII
POPULATION_SIZE = ${POP}
MAX_EVALUATIONS = ${MEVAL}
PROCESSORS = 1
PLOT_PARETO_FRONT = false
PYTHON3_DIRECTORY = /usr/bin/python3
VERBOSE = true
JAVA=${JAVABIN}
"


# if [[ ${BSILENT} == 0 ]]; then
#     echo test;
# fi
# exit


if [[ ${DOEVO} == 1 ]]; then
    echo "1.2) Synthesise optimal controller using EvoChecker from ${PRJDIR}/02-designspace/${MDL}.pm ..."
    echo "${ECCFG}" >${ECDIR}/config.properties

    # run evochecker (add policy output? for the pareto front)
    export LD_LIBRARY_PATH=${PRISMDIR}/lib

    echo "Running EvoChecker on '${MDL}' for objective '${OBJ}' ..."
    cd ${ECDIR}

    touch ${PRJDIR}/03-paretofronts/${MDL}-evo.log
    if [[ ${BSILENT} == 0 ]]; then
	gnome-terminal -- sh -c "tail -f ${PRJDIR}/03-paretofronts/${MDL}-evo.log"
    fi
    
    ${JAVABIN} \
        -Dfile.encoding=UTF-8 \
        -classpath ${ECDIR}/target/classes:${HOME}/.m2/repository/org/hamcrest/hamcrest-core/1.3/hamcrest-core-1.3.jar:${HOME}/.m2/repository/org/antlr/antlr4/4.5/antlr4-4.5.jar:${HOME}/.m2/repository/org/antlr/antlr4-runtime/4.5/antlr4-runtime-4.5.jar:${HOME}/.m2/repository/org/abego/treelayout/org.abego.treelayout.core/1.0.1/org.abego.treelayout.core-1.0.1.jar:${HOME}/.m2/repository/org/antlr/antlr-runtime/3.5.2/antlr-runtime-3.5.2.jar:${HOME}/.m2/repository/org/antlr/ST4/4.0.8/ST4-4.0.8.jar:${HOME}/.m2/repository/org/mariuszgromada/math/MathParser.org-mXparser/4.4.2/MathParser.org-mXparser-4.4.2.jar:${HOME}/.m2/repository/org/uma/jmetal/4.5.2/jmetal-4.5.2.jar:${HOME}/.m2/repository/org/modelChecker/PrismAPI/1.0.16/PrismAPI-1.0.16.jar \
        evochecker.EvoChecker >"${PRJDIR}/03-paretofronts/${MDL}-evo.log" \
        || { echo "EvoChecker failed." && exit; };
    
    cd ${PRJDIR}

    cp ${ECDIR}/`sed -n -e '/Pareto Front:/ s/Pareto Front: \(.*$\)/\1/p' <03-paretofronts/${MDL}-evo.log` ${FRONT}
    cp ${ECDIR}/`sed -n -e '/Pareto Set:/ s/Pareto Set: \(.*$\)/\1/p' <03-paretofronts/${MDL}-evo.log` ${SET} 
fi

if [[ ${DOVIS} == 1 && -e "${FRONT}" ]]; then 
    echo "Visualise Pareto front ${FRONT} ..."
    # visualise policy set and Pareto front: run matplotlib
    if [[ ${BSILENT} == 1 ]]; then
	if [[ ${DIM} == 2 ]]; then
            python3 ${VISTOOLDIR}/plotFront2D-silent.py ${FRONT} &
	else
            python3 ${VISTOOLDIR}/plotFront3D-silent.py ${FRONT} &
	fi
    else 
	if [[ ${DIM} == 2 ]]; then
            python3 ${ECDIR}/scripts/plotFront2D.py ${FRONT} &
	else
            python3 ${ECDIR}/scripts/plotFront3D.py ${FRONT} &
	fi
    fi
fi

################################################################################
if [[ ${DOPRISM} == 1 ]]; then 
    echo "1.3) Extract Pareto-optimal solution no. ${SOL} (for testing purposes) ..."
    NAMES=`sed -n -e '1p' <${SET}`
    VALUES=`sed -n -e "$((${SOL}+2))p" <${SET}`

    if [[ -z ${VALUES} ]]; then
	echo -n "Error: Solution ${SOL} not in range. Argument N for -p or " 
	echo "script parameter SOL must be in 1..$((`wc -l <${SET}` - 2))!"
	exit 1
    fi 
    echo -e "Chose solution:\n${NAMES}\n${VALUES}\n"

    # replace parameters
    # declare -A PARAMS
    # IFS=' '
    read -ra VALUES <<< "${VALUES}"
    i=0
    # create sed script for instantiating parameters with values
    for n in ${NAMES}; do
        # PARAMS[$n]=${VALUES[$i]}
        SEDSCR="${SEDSCR}s/evolve \(int\|double\) \(${n}\)\(.*$\)/const \1 \2 = ${VALUES[$i]};/;" 
        i=$(($i+1))
    done

    echo "Instantiate parametric stochastic process (pDTMC/pMDP->DTMC/MDP) ..."
    sed -e "${SEDSCR}" <02-designspace/${MDL}.pm >04-process-opt/${MDL}.dtmc

    # extract transition matrix from DTMC/MDP (previously instantiated by evoChecker)
    # Note/hack: switch -mdp has to be used to tell PRISM to export event labels from a DTMC model
    ${PRISM} 04-process-opt/${MDL}.dtmc -exportmodel 04-process-opt/${MDL}.tra,sta,lab \
             -mdp >04-process-opt/${MDL}-prism.log
    # ${PRISM} 01-process/${MDL}.pm -exportmodel 04-process-opt/${MDL}.tra,sta,lab -mdp
fi

if [[ ${DOANA} == 1 ]]; then
    echo "1.4) Perform (quantitative) analyses of the selected policy (DTMC) ..."
    if [[ -f "04-process-opt/${MDL}.dtmc" ]]; then
        rm "${PRJDIR}/03-analysis/${MDL}-prism.log"
        for p in TASK_FINISH5 COUNT_HARZARDS \
                 MISHAP_FREEDOM_MINMAX MISHAP_FREEDOM_AVG \
                 MISHAP_FREEDOM_MINMAX_FROM_HS \
                 MISHAP_FREEDOM_MINMAX_FROM_HRW \
                 MISHAP_FREEDOM_MINMAX_FROM_HC \
                 MISHAP_CNDPR_MINMAX_HS MISHAP_CNDPR_AVG_HS \
                 MISHAP_CNDPR_MINMAX_HRW MISHAP_CNDPR_AVG_HRW \
                 MISHAP_CNDPR_MINMAX_HC MISHAP_CNDPR_AVG_HC;
        do
	    ${PRISM} 04-process-opt/${MDL}.dtmc 02-requirements/${PROPS} \
                     -property "${p}" \
                     >>03-analysis/${MDL}-prism.log
        done 
        sed -i -n -e '/\(Model checking:\|Result:\)/p' 03-analysis/${MDL}-prism.log
    else
	echo "File 04-process-opt/${MDL}.dtmc missing. Please, use $0 -p ... first!"
    fi
fi

if [[ ${DOVIS} == 1 ]]; then
    if [[ -e 04-process-opt/${MDL}.sta
        && -e 04-process-opt/${MDL}.lab && -e 04-process-opt/${MDL}.tra ]]; then 
        echo "Visualise selected policy (DTMC) ..."
        ${YAPTOOLDIR}/yap-debug.sh -o 05-controller/ --vis-ctr 04-process-opt/${MDL} --adv2dot # --mdp
        mv 04-process-opt/${MDL}.pdf 04-process-opt/${MDL}-proc.pdf
	if [[ ${BSILENT} == 0 ]]; then
            ${PDFVIEWER} 04-process-opt/${MDL}-proc.pdf &
	fi
        echo "Visualise controller part of the policy ..."
        ${YAPTOOLDIR}/yap-debug.sh -o 05-controller/ --vis-ctr 04-process-opt/${MDL} --ctr2dot
        mv 04-process-opt/${MDL}.pdf 04-process-opt/${MDL}-ctr.pdf
	if [[ ${BSILENT} == 0 ]]; then
            ${PDFVIEWER} 04-process-opt/${MDL}-ctr.pdf &
	fi
    else
        echo "No policy ${MDL} to visualise. Please, provide .sta/tra/lab files."
    fi
fi

################################################################################
# extract and generate safety controller implementation (with YAP)

if [[ ${DOYAPC} == 1 ]]; then
    echo "3.1) Generate controller executable (DFA, C#) ..."
    yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t "${PRJDIR}/05-controller/${TPL}.cs" \
         -e "${PRJDIR}/04-process-opt/${MDL}" \
         -o "${PRJDIR}/05-controller/${MDL}.cs" \
         --global-logging -l 2 -f csicobot \
         --simulate no --synthesise platformcode || { echo "YAP failed." && exit; };
    # --interface .yap

    echo "3.2) Generate controller executable (C++) ..."
    { yapp -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -t 05-controller/workcell-ctr-tpl.hpp \
         -e 04-process-opt/${MDL} \
         -o 05-controller/${MDL}.hpp \
         --global-logging -l 2 -f cpp -v 1 \
	 --simulate no --synthesise platformcode && echo "Done."; } \
	|| { echo "YAP failed." && exit; };
fi

################################################################################
# generate abstract traces for developing use/test cases for run-time verification
if [[ ${DOTRACES} == 1 ]]; then
    echo "Obtain abstract trace for use case-based testing from model (pMDP) ..."
    ${PRISM} 04-process-opt/${MDL}.dtmc 02-requirements/genusecases.props \
             -mdp -v -extraddinfo >04-process-opt/${MDL}-prism.log
    # Note: Unfortunately, PRISM won't export the action labels, such that
    # the simulator GUI has to be used and the witnesses for the use cases to be exported.
fi

################################################################################
# import and abstract controller transition matrix into an FSM used for test generation
if [[ ${DOTEST} == 1 ]]; then
    echo "2.1) Generate test reference model (FSM) ..."
    yapp --synthesise testrefmod \
         -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -e "${PRJDIR}/04-process-opt/${MDL}" \
         --global-logging -l 2 -f libfsmtest --simulate no -v 2 \
         -o "${PRJDIR}/06-tests/${MDL}.txt" \
        || { echo "YAP failed." && exit; };
    
    echo "2.2) Generate test reference model (PDF) ..."
    yapp --synthesise testrefmod \
         -m "${PRJDIR}/01-risk/${ACT}.yap" \
         -e "${PRJDIR}/04-process-opt/${MDL}" \
         --global-logging -l 2 -f dot --simulate no -v 1 \
         -o "${PRJDIR}/06-tests/${MDL}.txt" \
        || { echo "YAP failed." && exit; };
    dot -Tpdf "${PRJDIR}/06-tests/${MDL}.dot" -o "${PRJDIR}/06-tests/${MDL}.tmp.pdf"
    pdfcrop --noverbose "${PRJDIR}/06-tests/${MDL}.tmp.pdf" "${PRJDIR}/06-tests/${MDL}.pdf"
    
    echo "4.1) Generate test suite (requires libfsmtest) ..."
    ${LIBFSMTESTDIR}/build/generator/generator -h -cself -a0 -null null \
		    06-tests/${MDL}_suite-h-0 06-tests/${MDL} &>06-tests/${MDL}.log
    
    echo "4.2) Generate test harness (requires libfsmtest, g++) ..."
    g++ -o 05-controller/${MDL}_testbench -std=c++17 -Wall -I. \
    	-I${LIBFSMTESTDIR}/src/libfsmtest \
	    05-controller/${MDL}_wrapper.cpp 05-controller/harness.cpp \
        /usr/lib/libfsmtest.so -lc++ &>05-controller/${MDL}.log

    echo "5.1) Run tests and record verdicts ..."
    05-controller/${MDL}_testbench 06-tests/${MDL}_suite-h-0 \
		  06-tests/${MDL}_suite-h-0 &>06-tests/${MDL}_results-h-0.log
    
    echo "5.2) Count failed test runs (if any) ..."
    grep -c -i "FAIL" <06-tests/${MDL}_results-h-0.log 2>&1 | tee 06-tests/${MDL}_failures.log
	    
    echo "Cleaning intermediate files ..."
    rm 06-tests/${MDL}.{aux,tex,dot,zip,tmp.pdf}
    
    echo "Creating ZIP archive ..."
    zip -j 06-tests/${MDL}.zip 06-tests/${MDL}.{in,out,fsm,state,pdf} \
	06-tests/${MDL}_suite-h-0.{in,out,fsm,state,txt} 06-tests/${MDL}_results-h-0.log
fi
