#!/bin/bash
# set -x
################################################################################
# YAP Tools: YAP-to-PRISM Debug Helper 
# CC BY-NC-SA 4.0 by Mario Gleirscher
# See http://creativecommons.org/licenses/by-nc-sa/4.0/
#
# This script aids in debugging PRISM models with non-zero end components
#
# Requirements: awk dot grep mkdir prism sed sort uniq yapp
################################################################################

USAGE="
Usage: ${0} [-o DIR] [--help | --chk-ctr PREFIX | --debug FILE [ACTION] 
       		| --syn-ctr FILE PROP 
                | --vis-ctr FILE ] [--mdp|--dtmc|--ctr]

Purpose: This tool can help with processing PRISM's debugging 
	 output, e.g., for reducing non-zero end components or for
	 adjusting reward structures to better fit the application.

Options:
	--chk-ctr PREFIX PROPFILE
                         	import model PREFIX.{tra,sta,lab} into PRISM
		  		and check all properties from 
				file PROPFILE
        --clean                 clean debug files
       	--debug FILE [ACTION]	output debugging data for MDP in FILE.prism
       			        with label ACTION to look up in FILE.tra
        --gen-ctr-spc [YAP] [PRISM] [OUT] 
                                generate the controller space using a
                                YAP model and a PRISM template and writing
                                output into the file OUT
        --mdp                   output an adversary MDP (default: DTMC)
        -o DIR                  output directory (default: ./debug/)
	--open PREFIX		open PRISM GUI with model and properties in 
	       			the files PREFIX.prism and PREFIX.props
	--syn-ctr FILE PROP    	generate adversary for MDP from 
		       		FILE-adv.prism and property PROP
        --vis-ctr FILE          visualise policy/adversary
	--help			print this help

Recommended procedure:
 1. --debug     check model generated by YAP
 2. --syn-ctr   synthesise controller with PRISM 
    Note: optimal synthesis with max-properties only for reward structures
    that avoid non-zero end components (MECs, SCCs)
 3. --chk-ctr   check controller with PRISM
 4. --open      open controller in PRISM for further analysis (optional)
 5. The controller can be exercised by (i) creating a simulation path in PRISM and
    (ii) resolving the choice (enabled actions) shown in PRISM 
    for the current state by applying the choice made in the controller.

Debugging hint for option --debug:
 1. If no self-loops in .trew, omit all singletons in .scc.
 2. For SCCs with > 1 states, check whether transitions are 
    listed in .trew file(s).
 3. If yes, check labels in .tra file and/or in .pdf graph.
 4. Change 
    a. SCC transitions rewards to 0 s.t. only entering/existing 
       transitions remain with non-zero rewards.
    b. guards of transition rewards to eliminate rewarded SCC 
       transitions in .trew file(s).
"

function help () {
    echo ${1}
    echo "${USAGE}"
    exit ${2}
}

STATEINFO=0
LOGALL='/\(warn\|error\|fail\|fault\|mistake\|miss\)/Ip;/^!/,$p'
PROC_DONE="DONE" # "\e[70G[\e[32mDONE\e[0m]"
PRISM=/opt/prism-4.5-linux64/bin/prism
PRISMGUI=/opt/prism-4.5-linux64/bin/xprism
OUTDIR=./debug/
YAPTOOLDIR=/usr/share/yapp/tools
MAXSTATES=5000

# check and initialize CLI options
OPTERR=0
while getopts "o:" OPTION; do
    case ${OPTION} in
	o) OUTDIR=${OPTARG} ;;
	*) # help
           ;;
    esac
done
shift 2
# shift $((OPTIND))

mkdir -p ${OUTDIR}

if [[ "${1}" == "--help" ]] || [[ -z "${1}" ]]; then
    help "" 0
elif [[ "${1}" == "--clean" ]]; then
    echo -n "Delete all files in ${OUTDIR} ..."
    rm -f ${OUTDIR}/*.*
    echo ${PROC_DONE}
    exit 0
elif [[ "${1}" == "--gen-ctr-spc" ]]; then
    if [[ -n "${2}" ]] && [[ -n "${3}" ]] && [[ -n "${4}" ]]; then
        yapp -m ${2}.yap -t ${3} -o ${4} --mark-source-code \
             -f prism -d multi-event-concurrent --simulate no --synthesise controller -l 2
        # --nosimplify 
        # --global-logging 
        exit 0
    else
        help "" 1
    fi
elif [[ "${1}" == "--syn-ctr" ]]; then
    if [[ -n "${2}" ]] && [[ -n "${3}" ]]; then
        if [[ -n "${4}" ]] && [[ "${4}" == "--mdp" ]]; then
            MDP=mdp
        fi
	BASE=$(basename -- "${2}")
	FILE="${OUTDIR}${BASE%.*}"
        echo "Synthesise adversary ..."
	${PRISM} ${2} -pctl "${3}" -exportadv${MDP} ${FILE}-adv.tra -s \
                 -exportstates ${FILE}-adv.sta -exportprodstates ${FILE}-adv.pst \
                 -exportlabels ${FILE}-adv.lab 
        # -exportmodel ${FILE}-adv.sta,lab 
        # -exportadvmdp -exportadv -exportunordered
        # -fair
        # cp ${FILE}-adv1.tra ${FILE}-adv.tra
	exit 0
    else
	help "Please, specify file name." 1
    fi
elif [[ "${1}" == "--vis-ctr" ]]; then
    if [[ -n "${2}" ]]; then
        if [[ `awk 'END {print NR}' ${2}.tra` -gt "${MAXSTATES}" ]]; then 
            echo -e "Adversary visualisation omitted because #states (= #transitions) too high."
            exit 0;
        elif [[ ! -a "${2}.tra" ]]; then
            echo -e "File ${2}.tra not found.";
            exit 0;
        fi
        echo "Visualise adversary ..."
        DOTFILE=${2}.dot
        echo -e "digraph P {\nnode [shape=circle,style=filled,color=lightgray];" > ${DOTFILE}
        if [[ -z "${3}" ]] || [[ "${3}" == "--adv2dot" ]]; then # complete policy of the plant
	    echo "Conversion of whole DTMC/MDP to DOT ..."
            sed -E -n -e '/^[0-9 ]+$/s/^([0-9]+)(\s+([0-9]+))?\s+([0-9]+)\s+([0-9\.]+)\s+$/\1 -> \4 [label="--:\5",color=gray];/p' \
                < ${2}.tra >> ${DOTFILE} # anonymous transitions
            sed -E -n -e '/hi?_/s/^([0-9]+)(\s+([0-9]+))?\s+([0-9]+)\s+([0-9\.]+)\s+(\S+)?$/\1 -> \4 [label="\6:\5"];/p' \
                < ${2}.tra >> ${DOTFILE} # operator events
            sed -E -n -e '/p_/s/^([0-9]+)(\s+([0-9]+))?\s+([0-9]+)\s+([0-9\.]+)\s+(\S+)?$/\1 -> \4 [label="\6:\5",color=blue];/p' \
                < ${2}.tra >> ${DOTFILE} # process/helper events
            sed -E -n -e '/(r|rw|wc?)_/s/^([0-9]+)(\s+([0-9]+))?\s+([0-9]+)\s+([0-9\.]+)\s+(\S+)?$/\1 -> \4 [label="\6:\5",color=red];/p' \
                < ${2}.tra >> ${DOTFILE} # plant events
            sed -E -n -e '/si?_/s/^([0-9]+)(\s+([0-9]+))?\s+([0-9]+)\s+([0-9\.]+)\s+(\S+)?$/\1 -> \4 [label="\6:\5",color=green];/p' \
                < ${2}.tra >> ${DOTFILE} # controller events
            if [[ -e "${2/-adv*/-adv}.pst" ]]; then
                # add state labels
                . ${YAPTOOLDIR}/gen-label-colors.sh ${2/-adv*/-adv} >>${DOTFILE}
            fi
            # add state variables
            if [[ ${STATEINFO} -ne 0 ]]; then
                sed -E -n -e '2,$ s/^([0-9]+):\((.+)\)$/\1 [label="\1\\n\2"];/p' \
                    <${2/-adv*/-adv.pst} >> ${DOTFILE}
            fi
        elif [[ "${3}" == "--ctr2dot" ]]; then # projection of policy to the safety controller
	    echo "Conversion of controller (DTMC/MDP) to DOT ..."
	    # format: srcst option? tgtst prob eventlbl?
            sed -E -n -e '/si?_/s/^([0-9]+)(\s+([0-9]+))?\s+([0-9]+)\s+([0-9\.]+)\s+(\S+)?$/\1 -> \4 [label="\6:\5",color=green];/p' \
                <${2}.tra >>${DOTFILE}
            # add state information
            if [[ ${STATEINFO} -ne 0 ]]; then
		# CAUTION: state indices of policy DTMC are from the tuple space DRAxMDP
		# not from the plain state space of the MDP, hence use adv.pst and not adv.sta!
                sed -E -n -e '2,$ s/^([0-9]+):\((.+)\)$/\1 [label="\1\\n\2"];/p' \
                    <${2/-adv*/-adv.pst} >>${DOTFILE}
            fi
        elif [[ "${3}" == "--mdp" ]]; then # obsoleted
	    echo "Conversion of MDP to DOT ..."
            sed -E -n -e '2,$ s/^([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9\.]+)\s+(\S+)\s*$/\1 -> \3 [label="\5:\4"];/p' \
                <${2}.tra >>${DOTFILE} 
	    # ${PRISM} -importstates ${2}.sta -importtrans ${2}.tra -dtmc -exporttransdotstates ${2}.dot
            # -importlabels ${FILE}-adv.lab -exporttransdotstates ${FILE}-adv.dot -mdp -dtmc
        fi
        echo -e "}" >> ${DOTFILE}
	dot -Tpdf ${2}.dot > ${2}.pdf
	echo "Visualisation done."
        exit 0
    else
	help "Please, specify file name." 1
    fi
elif [[ "${1}" == "--chk-ctr" ]]; then
    if [[ -n "${2}" ]] && [[ -n "${3}" ]]; then
        echo "Model checking the controller (the policy generated by PRISM) ..."
	${PRISM} -importstates ${2/-adv*/-adv.sta} -importlabels ${2/-adv*/-adv.lab} \
                 -importtrans ${2}.tra -dtmc ${3} -gs >${2/-adv/-checks-raw}.txt
        # -mdp -prop CTR_LIV -valiter
        grep -e 'Model checking:' -e 'Result:' -e 'Error:' < ${2/-adv/-checks-raw}.txt > ${2/-adv/-checks}.txt
	exit 0
    else
	help "Please, specify file name." 1
    fi
elif [[ "${1}" == "--open" ]]; then
    if [[ -n "${2}" ]]; then 
	BASE=$(basename -- "${2}")
	FILE="${MODDIR}${BASE%.*}"
	${PRISMGUI} ${FILE}.prism ${FILE}.props
	exit 0
    else
	help "Please, specify file name." 1
    fi
elif [[ "${1}" == "--debug" ]]; then
    BASE=$(basename -- "${2}")
    FILE="${BASE%.*}"
    EXT="${2##*.}"

    echo "Generating state list/transition matrix (.sta,.tra) with rewards (.srew,.trew) ..."
    ${PRISM} ${2} -exportmodel "${OUTDIR}${FILE}.sta,tra,srew,trew" | sed -n ${LOGALL}
    echo -e ${PROC_DONE}

    echo "Generating labelled transition graph (.dot,.pdf) ..."
    ${PRISM} ${2} -exportmodel "${OUTDIR}${FILE}.dot" | sed -n ${LOGALL}
    if [[ `awk 'END {print NR}' ${OUTDIR}${FILE}.dot` -lt "5000" ]]; then # wc -l produces verbose comments
        dot -Tpdf "${OUTDIR}${FILE}.dot" >"${OUTDIR}${FILE}.pdf" | sed -n ${LOGALL} 
        echo -e ${PROC_DONE}
    else
        echo -e "DOT-to-PDF omitted because of two many states and transitions."
    fi

    echo -n "Generating strongly connected components (SCC, .scc) ..."
    ${PRISM} ${2} -exportsccs stdout >"${OUTDIR}${FILE}.scc" | sed -n ${LOGALL}
    echo -e ${PROC_DONE}
    echo -n "Generating maximal end components (MEC, .mec) ..."
    ${PRISM} ${2} -exportmecs stdout >"${OUTDIR}${FILE}.mec" | sed -n ${LOGALL}
    echo -e ${PROC_DONE}
    echo -n "Generating bottom strongly connected components (BSCC, .bsc) ..."
    ${PRISM} ${2} -exportbsccs stdout >"${OUTDIR}${FILE}.bsc1" | sed -n ${LOGALL}
    echo -e ${PROC_DONE}

    echo -n "Generating labels (.lab) ..."
    ${PRISM} ${2} -exportlabels stdout >"${OUTDIR}${FILE}.lab" | sed -n ${LOGALL}
    echo -e ${PROC_DONE}
    
    echo "Checking .trew for self-loops (SCCs with only one state) ..."
    for i in ${OUTDIR}${FILE}*.trew; do
	echo -e "\t* $i :"
	sed -E -n '/^([0-9]+)\s+[0-9]+\s+\1\s+/p' <"$i" >${OUTDIR}${FILE}-debug.txt
    done;
    echo -e ${PROC_DONE}

    # echo "Listing SCCs with > 1 states:"
    # sed -E -n '/^SCC/,/^\s*$/p' "${FILE}.scc"
    for type in SCC MEC; do
	awk -f yap-debug.awk -v type=${type} "${OUTDIR}${FILE}.${type,,}" >>${OUTDIR}${FILE}-debug.txt
    done
    echo -e ${PROC_DONE}

    echo "Listing used transition labels ..."
    sed -E -n 's/^[0-9 ]+([[:alpha:]].*)\s*$/\1/p' <"${OUTDIR}${FILE}.tra" \
        | sort | uniq >>${OUTDIR}${FILE}-debug.txt
    echo -e ${PROC_DONE}

    if [[ -n "${3}" ]]; then
	echo "Checking .trew for transitions labelled with '${3}' ..."
	echo "(Format: <from> <idx> <to> <prob> [<label>])"
	# for each line in .tra ...
	for f in ${OUTDIR}${FILE}*.trew; do
	    echo "Check ${f} ..."
	    IFS=$'\n'
	    for tr in `sed -E -n "/\\s${3}($|\\s)/ {s/^([0-9]+\s+[0-9]+\s+[0-9]+).*$/\1/p}" "${OUTDIR}${FILE}.tra"`; do
		sed -E -n "/${tr}/p" "${f}" >>${OUTDIR}${FILE}-debug.txt
	    done 
	done
	echo -e ${PROC_DONE}
    fi

    echo -e "Debug output written to file ${OUTDIR}${FILE}-debug.txt."
fi
