;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; yap-mode - an Emacs major mode for YAP 0.8.1+20220513
;; CC BY-NC-SA 4.0 by Mario Gleirscher
;; see http://creativecommons.org/licenses/by-nc-sa/4.0/
;;
;; Thanks to ErgoEmacs.org for many practical eLisp hints.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (require 'ido)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Add the following 2 lines to your .emacs to activate YAP mode
;; for .yap files:
;;
;; (autoload 'yap-mode "/usr/share/yapp/yapp.el" "Major mode for YAP scripts" t)
;; (add-to-list 'auto-mode-alist '("\\.yap\\'" . yap-mode))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(setq yap-path "/usr/share/yapp/")

(setq yap-viewer-pdf "/usr/bin/zathura")
(setq yap-viewer-png "/usr/bin/eog")
(setq yap-viewer-svg "/usr/bin/eog")
(setq yap-format "latex") 
(setq dot-format "dot2tex") 
(setq img-format "svg") ;; only for sfdp and neato
(setq yap-outdir "output/")
(setq yap-verbosity-level 0)
(setq yap-calc-severity 0)
(setq yap-logging-level 2)
(setq yap-logging-global 1)
(setq yap-latex-fontsize "\\Large")
(setq yap-simplify 1)
(setq yap-expand-formulas 1)
(setq yap-logfile-global "~/.yap/yap.log") ;; tail -f
;; (setq yap-bin (concat yap-path "yapp"))
(setq yap-templatefile ".") ;; "input.txt"
(setq yap-bin (concat "java -jar " yap-path "yapp.jar --global-logging"))
(setq neato-bin (format "neato -x -T%s" img-format))
(setq sfdp-bin
      ;; (format "dot -Ksfdp -T%s" img-format)
      (format "sfdp -x -Goverlap=scale -T%s" img-format))
(setq dot2tex-bin (format "dot2tex --autosize --figpreamble=\"%s\" -c -f tikz -t raw"
                          yap-latex-fontsize))
(setq latex-bin "pdflatex -interaction=nonstopmode -output-directory=")
(setq yap-manual "/usr/share/yapp/doc/manual.pdf")
(setq yap-outputb "*yap output for %s*")
(setq cmd-clean (format "find %s -iregex \".*\\.\\(tmp\\.pdf\\|log\\|aux\\)\" -delete" yap-outdir))
(setq cmd-clean-all (format "find %s -iregex \".*\\.\\(tmp\\.pdf\\|dot\\|log\\|tex\\|aux\\)\" -delete" yap-outdir)) 

;; Some helper functions
(defun yap-display-version ()
  "Display version number of installed YAP (mode) version"
    (interactive)
    (message "You are using yap-mode 0.8.1+20220513."))

(defun yap-display-files (files)
  (let* ((process-connection-type nil))
    (message "%s" files)
    (dolist (f files)
      (when (file-exists-p f)
        (cond ((equal (file-name-extension f) "pdf") (start-process "" nil yap-viewer-pdf f))
              ((equal (file-name-extension f) "png") (start-process "" nil yap-viewer-png f))
              ((equal (file-name-extension f) "svg") (start-process "" nil yap-viewer-svg f)))))))

(defun yap-filetype (format)
  (let* ((process-connection-type nil))
    (cond ((or (equal format "evochecker") (equal format "prism")) '".pm")
          ((format "latex") '".tex"))))

(defun yap-clean-files ()
  (shell-command cmd-clean nil)) ;; integrate cmd-clean-all

(defun yap-save ()
  (when (not (buffer-file-name)) (save-buffer))
  (when (buffer-modified-p) (save-buffer)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; YAP commands and key bindings
;; inspired by: http://ergoemacs.org/emacs/elisp_run_current_file.html
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun yap-display-manual ()
  "Display PDF manual"
  (interactive)
  (progn
    (message "Display PDF manual ...")
    (yap-display-files (list yap-manual))))

(defun yap-display-settings ()
  "Display current settings"
  (interactive)
  (progn
    (message (concat "Current YAP settings:"
                     "\nSimplify risk space:\t\t%s"
                     "\nOutput format (for neato, sfdp):\t\t%s"
                     "\nDOT rendering engine:\t%s"
                     "\nDOT output format:\t%s"
                     "\nLogging level:\t\t%s"
                     "\nVerbosity level:\t%s"
                     "\nPDF viewer:\t\t%s"
                     "\nModel template:\t\t%s")
             yap-simplify yap-format dot-format img-format yap-logging-level
             yap-verbosity-level yap-viewer-pdf yap-templatefile)))

(defun yap-change-settings ()
  "Prompt user to adjust several yap-mode settings."
  (interactive)
  (progn
    (setq yap-format (completing-read "YAP output format: "
                      '("latex" "plain"
                        "evochecker" "prism" "libfsmtest" "prism" "pdtmcalldps"
                        "cpp" "cpp_lean" "csicobot")))
    (setq dot-format (completing-read
                      "DOT rendering engine: " '("dot2tex" "neato" "sfdp")))
    (setq img-format (completing-read
                      "DOT output format: " '("svg" "png")))
    (setq yap-logging-level (completing-read
                      "Logging level (0..no logging, 1..only SEVERE, 3..ALL): " '("2" "0" "1" "3")))
    (setq yap-logging-global (completing-read
                      "Global logging? (1..yes, 0..no): " '("1" "0")))
    (setq yap-simplify (completing-read
                      "Risk model: Simplify risk space? (1..yes, 0..no): " '("1" "0")))
    (setq yap-calc-severity (completing-read
                      "Risk model: Calculate severity (1..yes, 0..no): " '(("0" 0) ("1" 1))))
    (setq yap-expand-formulas (completing-read
                      "Synthesis: Expand formulas? (1..yes, 0..no): " '("1" "0")))
    (setq yap-verbosity-level (completing-read
                      "Verbose output (0..no details, >0..more details): " '(("0" 0) ("1" 1) ("2" 2) ("3" 3))))
    (setq yap-viewer-pdf (completing-read
                          "PDF viewer: " '("/usr/bin/zathura" "/usr/bin/evince" "/usr/bin/okular")))
    (setq yap-templatefile (read-file-name "External model template: "))
    (setq yap-outdir (read-file-name "Output directory: "))))

(defun yap-display-log ()
  "Display logs for current file"
  (interactive)
  (let* ((model (file-name-base buffer-file-name))
	 (logfile (if yap-logging-global
		      yap-logfile-global
		    (format "%s.yap.log" model)))
	 (logb (format "*yap log for %s*" model)))
    (setq cmd-yap-log (format "touch %s; tail -n +1 -f %s &" logfile logfile))
    (progn
      (message "Display logs for \"%s\" ..." model)
      (shell-command cmd-yap-log logb))))

(defun yap-display-model ()
  "Generate risk structure for current YAP file"
  (interactive)
  (yap-save)
  (let* ((resize-mini-windows nil)
         (modelfile (file-name-base buffer-file-name)))
    (setq cmd-yap (format "%s -v %s -l %s -m %s.yap --showmodel raw"
                          yap-bin yap-verbosity-level yap-logging-level modelfile))
    (progn
      (message "Displaying model for \"%s\" ..." modelfile)
      (shell-command
       (concat cmd-yap " || echo \"Exit with error code: ${?}\" &")
       (format yap-outputb (concat modelfile "-model"))))))

(defun yap-display-mincs ()
  "Generate all minimal cut sequences"
  (interactive)
  (yap-save)
  (let* ((resize-mini-windows nil)
         (modelfile (file-name-base buffer-file-name)))
    (setq cmd-yap (format "%s -v %s -l %s -m %s.yap --mincs -f emacs"
                          yap-bin yap-verbosity-level yap-logging-level modelfile))
    (progn
      (message "Displaying cut sequences for \"%s\" ..." modelfile)
      (shell-command
       (concat cmd-yap " || echo \"Exit with error code: ${?}\" &")
       (format yap-outputb (concat modelfile "-model"))))))

(defun yap-simulate-path ()
  "Simulate a random path through the activity graph"
  (interactive)
  (yap-save)
  (let* ((resize-mini-windows nil)
         (modelfile (file-name-base buffer-file-name))
	 (outputfile (concat yap-outdir modelfile "-rs")))
    (setq cmd-yap (format "%s -f latex --simulate random -v %s -l %s -m %s.yap -o %s.dot %s"
                          yap-bin yap-verbosity-level yap-logging-level modelfile outputfile
                          (if (equal yap-simplify "1") "" "--nosimplify")))
    (setq cmd-multidot (format "(for i in %s*.dot; do %s -o ${i/.dot/.tex} $i; done;)" yap-outdir yap-dot2tex-bin))
    (setq cmd-multipdf (format "(for i in %s*.tex; do %s%s $i; done;)" yap-outdir latex-bin yap-outdir))
    (progn
      (message "Simulate random path from \"%s\" (generating PDFs) ..." modelfile)
      (shell-command
       (concat "(" cmd-yap " && " cmd-multidot " && " cmd-multipdf " && " cmd-clean ") || echo \"Exit with error code: ${?}\" &")
       (format yap-outputb (concat modelfile "-rs"))))))

(defun yap-generate-activity-graph ()
  "Generate graph of activities reachable from the current activity"
  (interactive)
  (yap-save)
  (let* ((resize-mini-windows nil)
         (modelfile (file-name-base buffer-file-name))
	 (outputfile (concat yap-outdir modelfile "-act")))
    (setq cmd-yap (format "%s -v %s -l %s -m %s.yap --showmodel activities -o %s.dot" yap-bin
                          yap-verbosity-level yap-logging-level modelfile outputfile))
    (setq cmd-dotpdf (format "dot -Tpdf %s.dot -o %s.tmp.pdf" outputfile outputfile))
    (setq cmd-postproc (format "pdfcrop --noverbose %s.tmp.pdf %s.pdf" outputfile outputfile))
    (progn
      (message "Generating activity graph for \"%s\" (generating PDF) ..." modelfile)
      (shell-command
       (concat "(" cmd-yap " && " cmd-dotpdf " && " cmd-postproc " && " cmd-clean ") || echo \"Exit with error code: ${?}\" &")
       (format yap-outputb (concat modelfile "-act"))))))

(defun yap-generate-fault-tree ()
  "Generate fault tree from the current activity"
  (interactive)
  (yap-save)
  (let* ((resize-mini-windows nil)
         (modelfile (file-name-base buffer-file-name))
	 (outputfile (concat yap-outdir modelfile "-dft")))
    (setq cmd-yap (format "%s %s -m %s.yap --showmodel dft -o %s.dot" yap-bin
                          (if (equal yap-simplify "1") "" "--nosimplify")
                          modelfile outputfile))
    (setq cmd-dotpdf (format "dot -Tpdf %s.dot -o%s.tmp.pdf" outputfile outputfile))
    (setq cmd-postproc (format "pdfcrop --noverbose %s.tmp.pdf %s.pdf" outputfile outputfile))
    (progn
      (message "Generating fault tree for \"%s\" (generating PDF) ..." modelfile)
      (shell-command
       (concat "(" cmd-yap " && " cmd-dotpdf " && " cmd-postproc " && " cmd-clean ") || echo \"Exit with error code: ${?}\" &")
       (format yap-outputb (concat modelfile "-dft"))))))

(defun yap-generate-controller-model ()
  "Generate model for controller synthesis"
  (interactive)
  (yap-save)
  (let* ((modelfile (file-name-base buffer-file-name))
         (outputfile (concat yap-outdir
                             (if (equal (file-name-directory yap-outdir) yap-outdir)
                                 (modelfile (yap-filetype yap-format)) ""))))
    (setq cmd-yap
          (format "%s -l %s -m %s.yap --synthesise controller -d turn-based -f %s -t %s -o %s %s %s"
                  yap-bin yap-logging-level modelfile yap-format yap-templatefile outputfile
                  (if (equal yap-simplify "1") "" "--nosimplify")
                  (if (equal yap-expand-formulas "1") "" "--expand-formulas")))
    (progn
      (message "Generating controller model for \"%s\" ..." modelfile)
      (shell-command cmd-yap yap-outputb))))

(defun yap-explore-risk ()
  "Generate risk structure for current YAP file"
  (interactive)
  (yap-save)
  (let* ((resize-mini-windows nil)
         (modelfile (file-name-base buffer-file-name))
	 (outputfile (concat yap-outdir modelfile "-rs")))
    (setq cmd-yap
          (format "%s %s %s -m %s.yap -f %s -v %s -l %s --simulate initial -o %s.dot"
                  yap-bin
                  (if (equal yap-calc-severity "1") "--severity" "")
                  (if (equal yap-simplify "1") "" "--nosimplify")
                  modelfile yap-format yap-verbosity-level yap-logging-level outputfile))
    (cond ((equal dot-format "sfdp")
           (setq cmd-dot (format "%s -o %s.%s %s.dot" sfdp-bin outputfile img-format outputfile)))
          ((equal dot-format "neato")
           (setq cmd-dot (format "%s -o %s.%s %s.dot" neato-bin outputfile img-format outputfile)))
          ((equal dot-format "dot2tex")
           (setq cmd-dot (format "%s -o %s.tex %s.dot" dot2tex-bin outputfile outputfile))))
    (if (equal dot-format "dot2tex")
        (setq cmd-pdf (format "%s%s %s.tex" latex-bin yap-outdir outputfile))
      (setq cmd-pdf "true"))
    (progn
      (message "Exploring activity \"%s\" (generating PDF) ..." modelfile)
      (shell-command
       (concat "set -e; (echo 'Running YAP ...'"
               " && (" cmd-yap
               ") && echo 'Processing DOT file(s) ...'"
               " && (" cmd-dot
               ") && echo 'Converting to PDF/PNG/SVG ...'"
               " && (" cmd-pdf
               ") && echo 'Cleaning up ... '"
               " && (" cmd-clean
               ") && echo 'Done.') || (echo \"Exit with error code: ${?}\"; exit) &")
       (format yap-outputb (concat modelfile "-rs"))))))

(defun yap-dot-to-pdf ()
  "Generate PDF from risk structure"
  (interactive)
  (yap-save)
  (let* ((resize-mini-windows nil)
         (modelfile (file-name-base buffer-file-name))
	 (outputfile (concat yap-outdir modelfile)))
    (progn
      (message "Generating PDF from %s ..." buffer-file-name)
      (shell-command
       (format "dot -Tpdf -o %s.pdf %s.dot &" outputfile outputfile)
       (format yap-outputb buffer-file-name)))))

(defun yap-display-graphs ()
  "Display risk graph for current activity"
  (interactive)
  (let* ((modelfile (file-name-base buffer-file-name))
	 (riskgraph (mapcar (function (lambda (x) (concat (format "%s%s-rs." yap-outdir modelfile) x)))
                            '("pdf" "png" "svg")))
	 (activitygraph (format "%s%s-act.pdf" yap-outdir modelfile))
	 (faultree (format "%s%s-dft.pdf" yap-outdir modelfile))
         (simruntable (format "%s%s-simrun.pdf" yap-outdir modelfile)))
    (progn
      (message "Display graph(s) for \"%s\" ..." modelfile)
      (yap-display-files (append riskgraph (list activitygraph faultree simruntable))))))

(defun yap-build ()
  "Build everything and clean redundant files"
  (interactive)
  (let* ((modelfile (file-name-base buffer-file-name)))
    (progn
      (message "Build everything for \"%s\" ..." modelfile)
      (yap-explore-risk)
      (yap-generate-activity-graph))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; YAP comment syntax
;; inspired by: http://ergoemacs.org/emacs/elisp_write_major_mode_index.html
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(setq yap-mode-syntax-table
      (let ( (synTable (make-syntax-table)))
        (modify-syntax-entry ?# "<" synTable) ;; # ...
        (modify-syntax-entry ?\n ">" synTable)
        (modify-syntax-entry ?\/ ". 124" synTable) ;; /* ... */ and // ...
        (modify-syntax-entry ?* ". 23b" synTable)
 	synTable))

(setq yap-font-lock-keywords
      (let* (
	     ;; define several category of keywords
	     (x-keywords
	      '(;; settings
                "settings" "outputDepth" "endangermentDepth" "mitigationDepth"
                "simulationLength" "suppressMishaps" "suppressEndangerments"
                "suppressMitigations" "suppressResumptions" "suppressStateLabel"
                "suppressMishapLabel" "suppressEndangermentsLabel"
                "suppressMitigationsLabel" "requiredVersion" "extFile"
                "graphDirection" "grayscale" "riskDiscount" "allFactorsDirect"
                ;; general
                "for" "alias" "desc" "description" "enum"
                ;; activities
                "OperationalSituation" "Activity" "Task"
                "enable" "include" "successor" "initialState"
                ;; control loop, modes
                "ControlLoop" "Controller" "Application" "Plant" "Process"
                "item" "asset" "device" "partOf"
                "poweredBy" "embodiedBy" "implementedBy"
                "mode" "event" "sync" "function" "capability" "update" "safetyfun" "target" "filter"
                "guard" "cause" "pre" "inv" "post" "constraint"
                "state" "type" "role" "class" "hasFunction"
                ;; factor model
                "FactorModel" "HazardModel" "offRepair" "direct" "activatedBy" "detectedBy"
                "mitigatedBy" "resumedBy" "alleviatedBy" "requires"
                "requiresOneOf" "requiresNOf" "requiresMit" "requiresOcc" "requiresNot"
                "causes" "permits" "excludes" "prevents"
                "preventsMit" "mitPreventsMit" "mitCausesMit" "final" "mishap" "accident" "incident"
                "triggeredBy" "impacts" "phaseinv"
                ;; synthesis
                "distances" "weights"))
	     (x-types
              '("FALLBACK" "START_MITIGATE" "SHUTDOWN" "PHASECHANGE" "ENDANGER" "MISHAP" "FAIL"
                "SHARE" "MITIGATE" "RESUME" "LIMIT" "NOTIFY" "PREVENT" "PREVENT_CRASH"
                "DEESCALATE" "M2H_HANDOVER"
                "UNSPECIFIC" "COMPUTING_UNIT" "CONTROL_LOOP" "EXTERNAL" "TCFD_SP_MODULE"
                "AGENT" "CONTROLLER" "PLANT" "HUMAN_OPERATOR"
                "SENSOR" "ACTUATOR" "OPTICAL" "INDUCTIVE" "FAIL_SAFE" "ALLEVIATE"
                "DOMAIN"))
	     (x-constants '())
	     ;; (x-events '(""))
	     ;; (x-functions '("[]"))

	     ;; generate regex string for each category of keywords
	     (x-keywords-regexp (regexp-opt x-keywords 'words))
	     (x-types-regexp (regexp-opt x-types 'words))
	     (x-constants-regexp "\\([0-9]+\\)") ;; '(regexp-opt x-constants 'words))
	     (x-events-regexp "\\([a-zA-Z]+\\)") ;; (regexp-opt x-events 'words))
	     ;; (x-functions-regexp (regexp-opt x-functions 'words))
	     )

        `((,x-keywords-regexp . font-lock-keyword-face)
          (,x-types-regexp . font-lock-type-face)
          (,x-constants-regexp . (1 font-lock-constant-face))
          (,x-events-regexp . (1 font-lock-builtin-face))
          ;; (,x-functions-regexp . font-lock-function-name-face)
          )))

;;;###autoload
;; c-mode, prog-mode, fundamental-mode
(define-derived-mode yap-mode c-mode "YAP"
  "Major mode for input to the YAP tool…"
  (setq font-lock-defaults '((yap-font-lock-keywords) nil t)) ;; t = case-insensitive syntax highlighting
  (set-syntax-table yap-mode-syntax-table)
  (linum-mode 1)
  (local-set-key (kbd "C-h m") 'yap-display-manual)
  (local-set-key (kbd "C-x C-a a") 'yap-generate-activity-graph)
  (local-set-key (kbd "C-x C-a b") 'yap-build)
  (local-set-key (kbd "C-x C-a c") 'yap-generate-controller-model)
  (local-set-key (kbd "C-x C-a f") 'yap-generate-fault-tree)
  (local-set-key (kbd "C-x C-a p") 'yap-simulate-path)
  (local-set-key (kbd "C-x C-a r") 'yap-explore-risk)
  (local-set-key (kbd "C-x C-a s") 'yap-change-settings)
  (local-set-key (kbd "C-x C-d c") 'yap-display-mincs)
  (local-set-key (kbd "C-x C-d g") 'yap-display-graphs) ;; either activity graph or risk graph
  (local-set-key (kbd "C-x C-d l") 'yap-display-log)
  (local-set-key (kbd "C-x C-d m") 'yap-display-model)
  (local-set-key (kbd "C-x C-d p") 'yap-dot-to-pdf)
  (local-set-key (kbd "C-x C-d s") 'yap-display-settings)
  (setq-local comment-start "/*")
  (setq-local comment-start-skip "/\\*+[ \t]*")
  (setq-local comment-end "*/")
  (setq-local comment-end-skip "[ \t]*\\*+/")
  (setq-default indent-tabs-mode nil)
  ;; (setq-default tab-width 2) ;; globally
  (setq tab-width 2)) ;; current buffer

;; (add-hook 'yap-mode-hook 'some-function)

(provide 'yap-mode)
